use std::{collections::BTreeMap, mem::transmute, vec};

use macroquad::{color::colors, prelude::*};
use serde::{Deserialize, Serialize};

use crate::{
	theme::{BG_DARK, BG_LIGHT},
	ui::text_button,
	util::{draw_rectangle_rec, screen_centered_rect},
	Globals,
};

#[derive(Debug, Clone, Copy, Deserialize, Serialize, PartialEq, Eq, PartialOrd, Ord)]
#[repr(u8)]
pub enum ActionId {
	Undo,
	Redo,
	Copy,
	Cut,
	Paste,
	EraseSelection,
	ToggleMenu,
	Save,
	StartSim,
	StopSim,
	StepSim,
	CycleGroup,
	CycleGroupRevMod,

	Eraser,
	Selection,
	Blueprints,
	NoTool,
	TileBlock,
	TileSilo,
	TileButton,
	TileIOSilo,
	TileFlipper,
	TileDigit,
	TileMarble,
	TileGroupWire,
	TileGroupArrow,
	TileGroupMirror,
	TileGroupMath,
	TileGroupCompare,
	// just like in C, because this way doesn't need more dependencies
	_EnumSize,
}

impl Default for Input {
	fn default() -> Self {
		use Button::*;
		let mut bindings = [(); ActionId::SIZE].map(|_| Vec::new());
		let mut bind_key = |action, mods, trigger| {
			bindings[action as usize].push(Binding {
				blocking_modifiers: Vec::new(),
				modifiers: mods,
				trigger,
			});
		};
		bind_key(ActionId::Undo, vec![LCtrl], Z);
		bind_key(ActionId::Redo, vec![LCtrl], Y);
		bind_key(ActionId::Redo, vec![LCtrl, LShift], Z);
		bind_key(ActionId::Copy, vec![LCtrl], C);
		bind_key(ActionId::Cut, vec![LCtrl], X);
		bind_key(ActionId::Paste, vec![LCtrl], V);
		bind_key(ActionId::EraseSelection, vec![], Backspace);
		bind_key(ActionId::ToggleMenu, vec![], Escape);
		bind_key(ActionId::Save, vec![LCtrl], S);
		bind_key(ActionId::StartSim, vec![], Enter);
		bind_key(ActionId::StopSim, vec![], Enter);
		bind_key(ActionId::StepSim, vec![], Space);
		bind_key(ActionId::CycleGroup, vec![], Tab);
		bind_key(ActionId::CycleGroupRevMod, vec![], LShift);
		bind_key(ActionId::Eraser, vec![], X);
		bind_key(ActionId::Blueprints, vec![LCtrl], B);
		bind_key(ActionId::Selection, vec![], B);
		bind_key(ActionId::TileBlock, vec![], Q);
		bind_key(ActionId::TileSilo, vec![], W);
		bind_key(ActionId::TileButton, vec![], E);
		bind_key(ActionId::TileIOSilo, vec![], R);
		bind_key(ActionId::TileFlipper, vec![], T);
		bind_key(ActionId::TileDigit, vec![], Y);
		bind_key(ActionId::TileMarble, vec![], A);
		bind_key(ActionId::TileGroupWire, vec![], S);
		bind_key(ActionId::TileGroupArrow, vec![], D);
		bind_key(ActionId::TileGroupMirror, vec![], F);
		bind_key(ActionId::TileGroupMath, vec![], G);
		bind_key(ActionId::TileGroupCompare, vec![], H);

		Self {
			action_bindings: bindings,
			action_states: [BindingState::Off; ActionId::SIZE],
			key_states: [BindingState::Off; Button::SIZE],
			editing_binding: None,
			in_text_edit: false,
		}
	}
}

#[derive(Clone, Copy, Debug, Default, PartialEq)]
enum BindingState {
	#[default]
	Off,
	Pressed,
	Held,
	Released,
}

type InputMap = BTreeMap<String, Vec<Binding>>;

#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(from = "InputMap", into = "InputMap")]
pub struct Input {
	action_bindings: [Vec<Binding>; ActionId::SIZE],
	action_states: [BindingState; ActionId::SIZE],
	key_states: [BindingState; Button::SIZE],
	editing_binding: Option<(ActionId, usize, BindingEdit)>,
	pub in_text_edit: bool,
}

#[derive(Clone, Debug)]
enum BindingEdit {
	Init,
	Adding(Binding),
	Releasing(Binding),
}

impl Input {
	pub fn draw_edit(&mut self, globals: &mut Globals, mut y: f32) {
		if self.editing_binding.is_some() {
			globals.mouse.clear();
		}

		let buttons_x = 220.;
		let binding_text_x = buttons_x + 135.;

		for action_index in 0..ActionId::SIZE {
			let action = ActionId::from_usize(action_index).unwrap();

			// if self.action_states[action_index] == BindingState::Pressed {
			// 	draw_rectangle(200, y, 20, 20, colors::LIMEGREEN);
			// }
			draw_text(&format!("{action:?}"), 16., y, 20., colors::ORANGE);
			for (binding_index, binding) in self.action_bindings[action_index].iter().enumerate() {
				if text_button(&globals.mouse, buttons_x, y, 80., "remove") {
					self.action_bindings[action_index].remove(binding_index);
					self.update_modifier_blocks();
					return;
				}
				if text_button(&globals.mouse, buttons_x + 85., y, 45., "edit") {
					self.editing_binding = Some((action, binding_index, BindingEdit::Init));
				}
				//
				let trigger = format!("{:?}", binding.trigger);
				let mut x = binding_text_x;
				draw_text(&trigger, x, y + 5., 20., colors::GREEN);
				x += 10. + measure_text(&trigger, None, 20, 1.).width;
				//
				let modifiers = format!("{:?}", binding.modifiers);
				draw_text(&modifiers, x, y + 5., 20., colors::BLUE);
				x += 10. + measure_text(&modifiers, None, 20, 1.).width;
				//
				let conflicts = conflicts(&self.action_bindings, binding, action);
				if !conflicts.is_empty() {
					let conflict_text = format!("also used by: {conflicts:?}");
					draw_text(&conflict_text, x, y + 5., 20., colors::ORANGE);
					x += 10. + measure_text(&conflict_text, None, 20, 1.).width
				}
				//
				if !binding.blocking_modifiers.is_empty() {
					let blocking_text = format!("not while: {:?}", binding.blocking_modifiers);
					draw_text(&blocking_text, x, y + 5., 20., colors::GRAY);
				}
				y += 32.;
			}
			if text_button(&globals.mouse, buttons_x, y, 130., "add binding") {
				self.editing_binding = Some((
					action,
					self.action_bindings[action_index].len(),
					BindingEdit::Init,
				));
			}
			y += 45.;
		}

		if let Some((action, binding_index, edit_state)) = &mut self.editing_binding {
			globals.mouse.update();
			let border = screen_centered_rect(408., 128.);
			draw_rectangle_rec(border, BG_LIGHT);
			let bounds = screen_centered_rect(400., 120.);
			draw_rectangle_rec(bounds, BG_DARK);
			let x = bounds.x;
			let y = bounds.y;
			draw_text(
				&format!("editing binding for {action:?}"),
				x + 10.,
				y + 5.,
				20.,
				colors::WHITE,
			);
			let y = y + 30.;
			let ok_btn_x = x + 10.;
			let ok_btn_y = y + 40.;
			let ok_btn_width = 80.;
			let ok_btn_rect = Rect::new(ok_btn_x, ok_btn_y, ok_btn_width, 30.);

			for key_index in 0..Button::SIZE {
				let key = Button::from_usize(key_index).unwrap();
				match edit_state {
					BindingEdit::Init => {
						if key.just_pressed() {
							*edit_state = BindingEdit::Adding(Binding {
								blocking_modifiers: Vec::new(),
								modifiers: Vec::new(),
								trigger: key,
							});
						}
					}
					BindingEdit::Adding(binding) => {
						if key.just_pressed() {
							if key != binding.trigger && !binding.modifiers.contains(&key) {
								binding.modifiers.push(binding.trigger);
								binding.trigger = key;
							}
						} else if key.released() {
							if let Some(i) = binding.modifiers.iter().position(|&k| k == key) {
								binding.modifiers.remove(i);
								binding.modifiers.push(binding.trigger);
								binding.trigger = key;
							}
							*edit_state = BindingEdit::Releasing(binding.clone());
						}
					}
					BindingEdit::Releasing(_binding) => {
						let clicking_ok =
							globals.mouse.is_over(ok_btn_rect) && key == Button::MouseLeft;
						if key.just_pressed() && !clicking_ok {
							*edit_state = BindingEdit::Adding(Binding {
								blocking_modifiers: Vec::new(),
								modifiers: Vec::new(),
								trigger: key,
							});
						}
					}
				}
			}

			if let BindingEdit::Adding(b) | BindingEdit::Releasing(b) = &edit_state {
				let colour = if matches!(edit_state, BindingEdit::Releasing(_)) {
					colors::GREEN
				} else {
					colors::ORANGE
				};
				let text = format!("{:?} + {:?}", b.modifiers, b.trigger);
				draw_text(&text, x + 5., y + 5., 20., colour);

				let conflicts = conflicts(&self.action_bindings, b, *action);
				if !conflicts.is_empty() {
					draw_text(
						&format!("conflicts: {conflicts:?}"),
						x + 200.,
						y + 40.,
						20.,
						colors::ORANGE,
					);
				}
			}
			if text_button(&globals.mouse, ok_btn_x, ok_btn_y, ok_btn_width, "ok") {
				if let BindingEdit::Releasing(binding) = edit_state {
					let binding_list = &mut self.action_bindings[*action as usize];
					if *binding_index < binding_list.len() {
						binding_list[*binding_index] = binding.clone();
					} else {
						binding_list.push(binding.clone());
					}
					self.editing_binding = None;
					self.update_modifier_blocks();
				}
			}
			if text_button(&globals.mouse, x + 100., y + 40., 80., "cancel") {
				self.editing_binding = None;
			}
		}
	}

	pub fn update(&mut self) {
		for i in 0..Button::SIZE {
			let button = Button::from_usize(i).unwrap();
			let state = &mut self.key_states[i];
			*state = if button.is_down() {
				match state {
					BindingState::Off | BindingState::Released => BindingState::Pressed,
					BindingState::Pressed | BindingState::Held => BindingState::Held,
				}
			} else {
				match state {
					BindingState::Off | BindingState::Released => BindingState::Off,
					BindingState::Pressed | BindingState::Held => BindingState::Released,
				}
			}
		}
		for i in 0..ActionId::SIZE {
			let bindings = &self.action_bindings[i];
			let mut is_active = BindingState::Off;
			if !self.in_text_edit {
				for binding in bindings {
					if binding
						.modifiers
						.iter()
						.all(|&m| self.key_states[m as usize] == BindingState::Held)
						&& binding
							.blocking_modifiers
							.iter()
							.all(|&m| self.key_states[m as usize] == BindingState::Off)
					{
						is_active = is_active.or(self.key_states[binding.trigger as usize]);
					}
				}
			}
			self.action_states[i] = is_active;
		}
		self.in_text_edit = false;
	}

	pub fn is_pressed(&self, action: ActionId) -> bool {
		self.action_states[action as usize] == BindingState::Pressed
	}

	pub fn is_held(&self, action: ActionId) -> bool {
		self.action_states[action as usize] == BindingState::Pressed
			|| self.action_states[action as usize] == BindingState::Held
	}

	pub fn is_released(&self, action: ActionId) -> bool {
		self.action_states[action as usize] == BindingState::Released
	}

	/// Must be called after any binding has changed.
	/// Ensures a binding "S" is not triggered by "Ctrl+S", if "Ctrl+S" is bound to something else.
	pub fn update_modifier_blocks(&mut self) {
		for i in 0..ActionId::SIZE {
			let bindings = &self.action_bindings[i];
			for binding_index in 0..bindings.len() {
				let binding = &self.action_bindings[i][binding_index];
				let mut blocking_mods = Vec::new();
				for i in 0..ActionId::SIZE {
					let other_bindings = &self.action_bindings[i];
					for other_binding in other_bindings {
						if other_binding.trigger == binding.trigger {
							for modifier in &other_binding.modifiers {
								if !blocking_mods.contains(modifier)
									&& !binding.modifiers.contains(modifier)
								{
									blocking_mods.push(*modifier);
								}
							}
						}
					}
				}
				self.action_bindings[i][binding_index].blocking_modifiers = blocking_mods;
			}
		}
	}
}

fn conflicts(
	bindings: &[Vec<Binding>; ActionId::SIZE],
	search: &Binding,
	skip: ActionId,
) -> Vec<ActionId> {
	let mut matches = Vec::new();

	for i in 0..ActionId::SIZE {
		if skip as usize == i {
			continue;
		}
		let bindings = &bindings[i];
		for binding in bindings {
			if binding == search {
				matches.push(ActionId::from_usize(i).unwrap());
				break;
			}
		}
	}

	matches
}

impl ActionId {
	pub const SIZE: usize = Self::_EnumSize as usize;

	fn from_usize(val: usize) -> Option<Self> {
		if val < Self::SIZE {
			Some(unsafe { transmute::<u8, Self>(val as u8) })
		} else {
			None
		}
	}
}

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
pub struct Binding {
	#[serde(skip)]
	blocking_modifiers: Vec<Button>,
	modifiers: Vec<Button>,
	trigger: Button,
}

impl BindingState {
	fn or(self, rhs: Self) -> Self {
		use BindingState::*;
		match (self, rhs) {
			(Off, other) => other,
			(other, Off) => other,
			(Held, _) => Held,
			(_, Held) => Held,
			(Pressed, Pressed) => Pressed,
			(Released, Released) => Released,
			(Pressed, Released) => Held,
			(Released, Pressed) => Held,
		}
	}
}

impl From<InputMap> for Input {
	fn from(map: InputMap) -> Self {
		let mut new = Self::default();
		for (action, loaded_bindings) in map {
			let temp_json = format!("\"{action}\"");
			if let Ok(action) = serde_json::from_str::<ActionId>(&temp_json) {
				new.action_bindings[action as usize] = loaded_bindings;
			} else {
				println!("'{action}' is not a valid action id, bindings discarded");
			}
		}
		new
	}
}

impl From<Input> for InputMap {
	fn from(value: Input) -> Self {
		value
			.action_bindings
			.iter()
			.enumerate()
			.map(|(i, b)| {
				// for this to panic, the .bindings array would have to be larger than ActionId::SIZE
				let action = ActionId::from_usize(i).unwrap();
				(format!("{action:?}"), b.clone())
			})
			.collect()
	}
}

#[derive(Debug, Clone, Copy, Serialize, Deserialize, PartialEq)]
#[repr(u8)]
enum Button {
	MouseLeft,
	MouseRight,
	MouseMiddle,
	// Mouse3,
	// Mouse4,
	// Mouse5,
	// Mouse6,
	Apostrophe,
	Comma,
	Minus,
	Period,
	Slash,
	Zero,
	One,
	Two,
	Three,
	Four,
	Five,
	Six,
	Seven,
	Eight,
	Nine,
	Semicolon,
	Equal,
	A,
	B,
	C,
	D,
	E,
	F,
	G,
	H,
	I,
	J,
	K,
	L,
	M,
	N,
	O,
	P,
	Q,
	R,
	S,
	T,
	U,
	V,
	W,
	X,
	Y,
	Z,
	LeftBracket,
	Backslash,
	RightBracket,
	Grave,
	Space,
	Escape,
	Enter,
	Tab,
	Backspace,
	Insert,
	Delete,
	Right,
	Left,
	Down,
	Up,
	PageUp,
	PageDown,
	Home,
	End,
	CapsLock,
	ScrollLock,
	NumLock,
	PrintScreen,
	Pause,
	F1,
	F2,
	F3,
	F4,
	F5,
	F6,
	F7,
	F8,
	F9,
	F10,
	F11,
	F12,
	LShift,
	LCtrl,
	LAlt,
	LeftSuper,
	RShift,
	RCtrl,
	RAlt,
	RightSuper,
	Menu,
	Kp0,
	Kp1,
	Kp2,
	Kp3,
	Kp4,
	Kp5,
	Kp6,
	Kp7,
	Kp8,
	Kp9,
	KpDecimal,
	KpDivide,
	KpMultiply,
	KpSubtract,
	KpAdd,
	KpEnter,
	KpEqual,
	Back,
	//
	_EnumSize,
}

enum RlInput {
	Key(KeyCode),
	Mouse(MouseButton),
}

impl Button {
	const SIZE: usize = Self::_EnumSize as usize;

	fn from_usize(val: usize) -> Option<Self> {
		if val < Self::SIZE {
			Some(unsafe { transmute::<u8, Self>(val as u8) })
		} else {
			None
		}
	}

	fn is_down(self) -> bool {
		match self.to_raylib() {
			RlInput::Key(key) => is_key_down(key),
			RlInput::Mouse(btn) => is_mouse_button_down(btn),
		}
	}

	fn just_pressed(self) -> bool {
		match self.to_raylib() {
			RlInput::Key(key) => is_key_pressed(key),
			RlInput::Mouse(btn) => is_mouse_button_pressed(btn),
		}
	}

	fn released(self) -> bool {
		match self.to_raylib() {
			RlInput::Key(key) => is_key_released(key),
			RlInput::Mouse(btn) => is_mouse_button_released(btn),
		}
	}

	fn to_raylib(self) -> RlInput {
		use RlInput::*;
		match self {
			Button::MouseLeft => Mouse(MouseButton::Left),
			Button::MouseRight => Mouse(MouseButton::Right),
			Button::MouseMiddle => Mouse(MouseButton::Middle),
			// Button::Mouse3 => Mouse(MouseButton::MOUSE_BUTTON_SIDE),
			// Button::Mouse4 => Mouse(MouseButton::MOUSE_BUTTON_EXTRA),
			// Button::Mouse5 => Mouse(MouseButton::MOUSE_BUTTON_FORWARD),
			// Button::Mouse6 => Mouse(MouseButton::MOUSE_BUTTON_BACK),
			Button::Apostrophe => Key(KeyCode::Apostrophe),
			Button::Comma => Key(KeyCode::Comma),
			Button::Minus => Key(KeyCode::Minus),
			Button::Period => Key(KeyCode::Period),
			Button::Slash => Key(KeyCode::Slash),
			Button::Zero => Key(KeyCode::Key0),
			Button::One => Key(KeyCode::Key1),
			Button::Two => Key(KeyCode::Key2),
			Button::Three => Key(KeyCode::Key3),
			Button::Four => Key(KeyCode::Key4),
			Button::Five => Key(KeyCode::Key5),
			Button::Six => Key(KeyCode::Key6),
			Button::Seven => Key(KeyCode::Key7),
			Button::Eight => Key(KeyCode::Key8),
			Button::Nine => Key(KeyCode::Key9),
			Button::Semicolon => Key(KeyCode::Semicolon),
			Button::Equal => Key(KeyCode::Equal),
			Button::A => Key(KeyCode::A),
			Button::B => Key(KeyCode::B),
			Button::C => Key(KeyCode::C),
			Button::D => Key(KeyCode::D),
			Button::E => Key(KeyCode::E),
			Button::F => Key(KeyCode::F),
			Button::G => Key(KeyCode::G),
			Button::H => Key(KeyCode::H),
			Button::I => Key(KeyCode::I),
			Button::J => Key(KeyCode::J),
			Button::K => Key(KeyCode::K),
			Button::L => Key(KeyCode::L),
			Button::M => Key(KeyCode::M),
			Button::N => Key(KeyCode::N),
			Button::O => Key(KeyCode::O),
			Button::P => Key(KeyCode::P),
			Button::Q => Key(KeyCode::Q),
			Button::R => Key(KeyCode::R),
			Button::S => Key(KeyCode::S),
			Button::T => Key(KeyCode::T),
			Button::U => Key(KeyCode::U),
			Button::V => Key(KeyCode::V),
			Button::W => Key(KeyCode::W),
			Button::X => Key(KeyCode::X),
			Button::Y => Key(KeyCode::Y),
			Button::Z => Key(KeyCode::Z),
			Button::LeftBracket => Key(KeyCode::LeftBracket),
			Button::Backslash => Key(KeyCode::Backslash),
			Button::RightBracket => Key(KeyCode::RightBracket),
			Button::Grave => Key(KeyCode::GraveAccent),
			Button::Space => Key(KeyCode::Space),
			Button::Escape => Key(KeyCode::Escape),
			Button::Enter => Key(KeyCode::Enter),
			Button::Tab => Key(KeyCode::Tab),
			Button::Backspace => Key(KeyCode::Backspace),
			Button::Insert => Key(KeyCode::Insert),
			Button::Delete => Key(KeyCode::Delete),
			Button::Right => Key(KeyCode::Right),
			Button::Left => Key(KeyCode::Left),
			Button::Down => Key(KeyCode::Down),
			Button::Up => Key(KeyCode::Up),
			Button::PageUp => Key(KeyCode::PageUp),
			Button::PageDown => Key(KeyCode::PageDown),
			Button::Home => Key(KeyCode::Home),
			Button::End => Key(KeyCode::End),
			Button::CapsLock => Key(KeyCode::CapsLock),
			Button::ScrollLock => Key(KeyCode::ScrollLock),
			Button::NumLock => Key(KeyCode::NumLock),
			Button::PrintScreen => Key(KeyCode::PrintScreen),
			Button::Pause => Key(KeyCode::Pause),
			Button::F1 => Key(KeyCode::F1),
			Button::F2 => Key(KeyCode::F2),
			Button::F3 => Key(KeyCode::F3),
			Button::F4 => Key(KeyCode::F4),
			Button::F5 => Key(KeyCode::F5),
			Button::F6 => Key(KeyCode::F6),
			Button::F7 => Key(KeyCode::F7),
			Button::F8 => Key(KeyCode::F8),
			Button::F9 => Key(KeyCode::F9),
			Button::F10 => Key(KeyCode::F10),
			Button::F11 => Key(KeyCode::F11),
			Button::F12 => Key(KeyCode::F12),
			Button::LShift => Key(KeyCode::LeftShift),
			Button::LCtrl => Key(KeyCode::LeftControl),
			Button::LAlt => Key(KeyCode::LeftAlt),
			Button::LeftSuper => Key(KeyCode::LeftSuper),
			Button::RShift => Key(KeyCode::RightShift),
			Button::RCtrl => Key(KeyCode::RightControl),
			Button::RAlt => Key(KeyCode::RightAlt),
			Button::RightSuper => Key(KeyCode::RightSuper),
			Button::Menu => Key(KeyCode::Menu),
			Button::Kp0 => Key(KeyCode::Kp0),
			Button::Kp1 => Key(KeyCode::Kp1),
			Button::Kp2 => Key(KeyCode::Kp2),
			Button::Kp3 => Key(KeyCode::Kp3),
			Button::Kp4 => Key(KeyCode::Kp4),
			Button::Kp5 => Key(KeyCode::Kp5),
			Button::Kp6 => Key(KeyCode::Kp6),
			Button::Kp7 => Key(KeyCode::Kp7),
			Button::Kp8 => Key(KeyCode::Kp8),
			Button::Kp9 => Key(KeyCode::Kp9),
			Button::KpDecimal => Key(KeyCode::KpDecimal),
			Button::KpDivide => Key(KeyCode::KpDivide),
			Button::KpMultiply => Key(KeyCode::KpMultiply),
			Button::KpSubtract => Key(KeyCode::KpSubtract),
			Button::KpAdd => Key(KeyCode::KpAdd),
			Button::KpEnter => Key(KeyCode::KpEnter),
			Button::KpEqual => Key(KeyCode::KpEqual),
			Button::Back => Key(KeyCode::Back),
			Button::_EnumSize => unreachable!(),
		}
	}
}
