use std::{
	collections::HashMap,
	fs::{create_dir_all, read_dir, read_to_string, File},
	io::Write,
};

use macroquad::{color::colors, miniquad::window::set_mouse_cursor, prelude::*};

use marble_machinations::*;

use config::Config;
use editor::{Editor, ExitState};
use level::{Chapter, Level};
use solution::Solution;
use theme::*;
use ui::{simple_option_button, tex32_button, text_button, text_input, ShapedText, Tooltip};
use util::*;

const TITLE_TEXT: &str = concat!("Marble Machinations v", env!("CARGO_PKG_VERSION"));

struct Game {
	chapters: Vec<Chapter>,
	level_scroll: f32,
	solutions: HashMap<String, Vec<Solution>>,
	open_editor: Option<Editor>,
	selected_level: (usize, usize),
	selected_solution: usize,
	delete_solution: Option<usize>,
	editing_solution_name: bool,
	level_desc_text: ShapedText,
	globals: Globals,
	edit_settings: Option<Config>,
}

#[macroquad::main("what does thsi doa")]
async fn main() {
	// let (mut rl, thread) = raylib::init().resizable().title(TITLE_TEXT).build();
	// set_target_fps(60);
	// set_window_min_size(640, 480);
	set_mouse_cursor(miniquad::CursorIcon::Crosshair);


	let mut game = Game::new().await;
	game.run().await
}

impl Game {
	async fn new() -> Self {
		let solutions = get_solutions();
		let chapters = get_chapters(&solutions);

		Self {
			chapters,
			level_scroll: 0.,
			solutions,
			open_editor: None,
			selected_level: (0, 0),
			selected_solution: 0,
			delete_solution: None,
			editing_solution_name: false,
			level_desc_text: ShapedText::new(20.),
			globals: Globals::new().await,
			edit_settings: None,
		}
	}

	async fn run(&mut self) {
		loop {
			self.globals.update();
			if let Some(editor) = &mut self.open_editor {
				editor.update(&mut self.globals);
				editor.draw(&mut self.globals);
				match editor.get_exit_state() {
					ExitState::Dont => (),
					ExitState::ExitAndSave => {
						let solution = &mut self.solutions.get_mut(editor.level_id()).unwrap()
							[self.selected_solution];
						solution.board = editor.source_board().clone();
						solution.score = editor.score();
						solution.save();
						self.open_editor = None;
					}
					ExitState::Save => {
						let solution = &mut self.solutions.get_mut(editor.level_id()).unwrap()
							[self.selected_solution];
						solution.board = editor.source_board().clone();
						solution.score = editor.score();
						solution.save();
					}
				}
			} else if let Some(config) = &mut self.edit_settings {
				clear_background(BG_DARK);
				match config.draw_edit(&mut self.globals) {
					config::MenuReturn::Stay => (),
					config::MenuReturn::StaySave => {
						self.globals.config = config.clone();
						self.save_config();
					}
					config::MenuReturn::ReturnSave => {
						self.globals.config = config.clone();
						self.save_config();
						self.edit_settings = None;
					}
					config::MenuReturn::ReturnCancel => self.edit_settings = None,
				}
			} else {
				self.draw();
			}
			next_frame().await
		}
	}

	fn draw(&mut self) {
		clear_background(BG_DARK);

		let mut tooltip = Tooltip::default();
		tooltip.init_frame();

		draw_text(
			TITLE_TEXT,
			screen_width() - 275.,
			screen_height() - 20.,
			20.,
			colors::GRAY,
		);

		let level_list_width = (screen_width() / 3.).min(400.);

		draw_rectangle(0., 0., level_list_width, screen_height(), BG_MEDIUM);

		if text_button(
			&self.globals.mouse,
			screen_width() - 100.,
			screen_height() - 70.,
			90.,
			"settings",
		) {
			self.edit_settings = Some(self.globals.config.clone());
		}

		const ENTRY_SPACING: f32 = 65.;
		let fit_on_screen = (screen_height() / ENTRY_SPACING) as usize;
		let max_scroll = self
			.chapters
			.iter()
			.map(|c| 1 + if c.visible { c.levels.len() } else { 0 })
			.sum::<usize>()
			.saturating_sub(fit_on_screen) as f32
			* ENTRY_SPACING;
		if self.globals.mouse.pos().x < level_list_width as f32 {
			if self.globals.mouse.scroll() == Some(Scroll::Down) && self.level_scroll < max_scroll {
				self.level_scroll += ENTRY_SPACING;
			}
			if self.globals.mouse.scroll() == Some(Scroll::Up) && self.level_scroll > 0. {
				self.level_scroll -= ENTRY_SPACING;
			}
		}

		let mut y = 10. - self.level_scroll;
		for (chapter_i, chapter) in self.chapters.iter_mut().enumerate() {
			let bounds = Rect::new(5., y - 5., level_list_width - 10., ENTRY_SPACING - 5.);
			draw_rectangle_rec(bounds, BG_DARK);
			draw_text(&chapter.title, 10., y, 30., FG_CHAPTER_TITLE);
			let subtitle = format!("{} levels", chapter.levels.len());
			draw_text(&subtitle, 10., y + 30., 20., colors::WHITE);
			y += ENTRY_SPACING;
			let clicked_this =
				self.globals.mouse.left_click() && self.globals.mouse.is_over(bounds);

			if clicked_this {
				chapter.visible = !chapter.visible;
			}

			if !chapter.visible {
				continue;
			}

			for (level_index, level) in chapter.levels.iter().enumerate() {
				let bounds = Rect::new(5., y - 5., level_list_width - 10., ENTRY_SPACING - 5.);
				let clicked_this =
					self.globals.mouse.left_click() && self.globals.mouse.is_over(bounds);

				if clicked_this && self.selected_level != (chapter_i, level_index) {
					self.editing_solution_name = false;
					self.selected_level = (chapter_i, level_index);
					self.selected_solution = 0;
					self.delete_solution = None;
					// select the last solution of the level, if there is one
					if let Some(solutions) = self.solutions.get(level.id()) {
						self.selected_solution = solutions.len().saturating_sub(1);
					}
				}
				draw_rectangle_rec(
					bounds,
					widget_bg(self.selected_level == (chapter_i, level_index)),
				);

				let mut title_color = colors::WHITE;
				if let Some(solutions) = self.solutions.get(level.id()) {
					if solutions.iter().any(|s| s.score.is_some()) {
						title_color = colors::GREEN;
					}
				}
				draw_text(level.name(), 10., y, 30., title_color);
				let solution_count = self
					.solutions
					.get(level.id())
					.map(Vec::len)
					.unwrap_or_default();
				let subtext = format!("solutions: {solution_count}");
				let subtext_color = if solution_count > 0 {
					colors::GOLD
				} else {
					colors::LIGHTGRAY
				};
				draw_text(&subtext, 10., y + 30., 20., subtext_color);
				y += ENTRY_SPACING;
			}
		}

		if let Some(level) = self
			.chapters
			.get(self.selected_level.0)
			.and_then(|c| c.levels.get(self.selected_level.1))
		{
			draw_text(level.name(), level_list_width + 10., 10., 40., colors::SKYBLUE);
			draw_text(level.id(), level_list_width + 10., 50., 10., colors::GRAY);

			let mut y = 70.;
			self.level_desc_text.set_text(level.description());
			self.level_desc_text
				.update_width(screen_width() - level_list_width - 30.);
			self.level_desc_text.draw(level_list_width + 10., y);
			y += self.level_desc_text.height() + 10.;

			if let Some(solutions) = self.solutions.get_mut(level.id()) {
				let solution_entry_height = 40.;
				let entry_width = 200.;
				let mut solution_y = y;
				for (solution_index, solution) in solutions.iter().enumerate() {
					if simple_option_button(
						&self.globals.mouse,
						Rect::new(
							level_list_width + 10.,
							solution_y,
							entry_width,
							solution_entry_height,
						),
						solution_index,
						&mut self.selected_solution,
					) {
						self.editing_solution_name = false;
					}
					let name_color = if solution.score.is_some() {
						colors::LIME
					} else {
						colors::ORANGE
					};
					draw_text(
						&solution.name,
						level_list_width + 15.,
						solution_y + 5.,
						20.,
						name_color,
					);
					draw_text(
						&solution.score_text(),
						level_list_width + 15.,
						solution_y + 25.,
						10.,
						colors::WHITE,
					);
					if tex32_button(
						 &self.globals.mouse,
						(level_list_width + entry_width + 15., solution_y + 4.),
						self.globals.get_tex("cancel"),
						(&mut tooltip, "delete"),
					) {
						self.delete_solution = Some(solution_index);
					}
					solution_y += solution_entry_height + 10.;
				}

				let next_id = get_free_id(solutions, Solution::id);
				if text_button(
					&self.globals.mouse,
					level_list_width + 10.,
					solution_y,
					entry_width,
					"new solution",
				) {
					self.selected_solution = solutions.len();
					solutions.push(Solution::new(level, next_id));
				}

				if let Some(i) = self.delete_solution {
					let text = format!("really delete solution '{}'?", &solutions[i].name);
					let y = (solution_y + 40.).max(240.);
					let x = level_list_width + 10.;
					draw_text(&text, x, y, 20., colors::ORANGE);
					if text_button(&self.globals.mouse, x, y + 30., 100., "yes") {
						solutions[i].remove_file();
						solutions.remove(i);
						self.delete_solution = None;
					}
					if text_button(&self.globals.mouse, x + 110., y + 30., 100., "no") {
						self.delete_solution = None;
					}
				}

				if let Some(solution) = solutions.get_mut(self.selected_solution) {
					let column_x = level_list_width + entry_width + 56.;
					let bounds = Rect::new(column_x as f32, y as f32, 220., 30.);
					if text_input(
						&mut self.globals,
						bounds,
						&mut solution.name,
						&mut self.editing_solution_name,
						24,
						true,
					) {
						solution.save();
					}
					let id_text = format!("{}", solution.id());
					draw_text(&id_text, column_x, y + 35., 10., colors::GRAY);

					if text_button(&self.globals.mouse, column_x, y + 50., 220., "clone") {
						let cloned = solution.new_copy(next_id);
						self.selected_solution = solutions.len();
						solutions.push(cloned);
						return;
					}

					if text_button(&self.globals.mouse, column_x, y + 85., 220., "edit") {
						let mut editor = Editor::new(solution.clone(), level.clone());
						editor.center_view();
						self.open_editor = Some(editor);
					}
				}
			} else {
				self.solutions.insert(level.id().to_owned(), Vec::new());
			}
		}
		tooltip.draw();
	}

	fn save_config(&self) {
		_ = create_dir_all(userdata_dir());
		let path = userdata_dir().join(CONFIG_FILE_NAME);
		let json = serde_json::to_string_pretty(&self.globals.config).unwrap();
		match File::create(path) {
			Ok(mut f) => {
				_ = f.write_all(json.as_bytes());
			}
			Err(e) => println!("error saving config: {e}"),
		}
	}
}

fn get_chapters(solutions: &HashMap<String, Vec<Solution>>) -> Vec<Chapter> {
	let mut chapters = Vec::<Chapter>::new();
	for d in read_dir("levels").unwrap().flatten() {
		if let Ok(text) = read_to_string(d.path()) {
			if let Ok(mut chapter) = serde_json::from_str::<Chapter>(&text) {
				chapter.visible = false;
				for level in &chapter.levels {
					if solutions
						.get(level.id())
						.map(|s| s.iter().any(|s| s.score.is_some()))
						!= Some(true)
					{
						// only expand chapters where at least one level has no complete solutions
						chapter.visible = true;
						break;
					}
				}
				chapters.push(chapter);
			}
		}
	}
	chapters.sort_unstable_by_key(|c| c.title.clone());

	// user levels
	let mut custom_levels = Vec::new();
	let custom_level_dir = userdata_dir().join("levels");
	if custom_level_dir.is_dir() {
		for d in read_dir(custom_level_dir).unwrap().flatten() {
			if let Ok(text) = read_to_string(d.path()) {
				if let Ok(level) = serde_json::from_str(&text) {
					custom_levels.push(level);
				}
			}
		}
	}
	chapters.push(Chapter {
		title: "Custom levels".into(),
		levels: custom_levels,
		visible: true,
	});
	if let Some(orphans) = find_orphans(&chapters, solutions) {
		chapters.push(orphans);
	}
	chapters
}

fn get_solutions() -> HashMap<String, Vec<Solution>> {
	let mut by_level = HashMap::new();
	let solution_dir = userdata_dir().join("solutions");
	if let Ok(dir_contents) = read_dir(solution_dir) {
		for dir in dir_contents.flatten() {
			if dir.path().is_dir() {
				let level_name = dir.file_name().to_string_lossy().to_string();
				let mut solutions = Vec::new();
				if let Ok(files) = read_dir(dir.path()) {
					for file in files.flatten() {
						let s = read_to_string(file.path())
							.ok()
							.as_deref()
							.and_then(|s| serde_json::from_str(s).ok());
						if let Some(solution) = s {
							solutions.push(solution);
						}
					}
					solutions.sort_unstable_by_key(Solution::id);
				}
				if !solutions.is_empty() {
					by_level.insert(level_name, solutions);
				}
			}
		}
	}
	by_level
}

fn find_orphans(
	chapters: &[Chapter],
	solutions: &HashMap<String, Vec<Solution>>,
) -> Option<Chapter> {
	let mut orphan_levels = Vec::new();
	'outer: for id in solutions.keys() {
		for c in chapters {
			for l in &c.levels {
				if l.id() == id {
					continue 'outer;
				}
			}
		}
		orphan_levels.push(Level::new_orphan(id))
	}
	if orphan_levels.is_empty() {
		None
	} else {
		Some(Chapter {
			title: "Missing levels".into(),
			levels: orphan_levels,
			visible: false,
		})
	}
}
