use std::ops::Range;

use crate::{
	theme::*,
	util::{draw_rectangle_rec, draw_scaled_texture, MouseInput, Scroll, Textures},
	Globals,
};
use macroquad::{color::colors, prelude::*};

#[derive(Debug)]
pub struct ShapedText {
	text: String,
	max_width: f32,
	font_size: f32,
	lines: Vec<Range<usize>>,
}

impl ShapedText {
	pub fn new(font_size: f32) -> Self {
		Self {
			text: String::new(),
			font_size,
			max_width: 500.,
			lines: Vec::new(),
		}
	}

	pub fn set_text(&mut self, text: &str) {
		if text != self.text {
			self.text = text.to_owned();
			self.lines.clear();
		}
	}

	pub fn height(&self) -> f32 {
		self.font_size * self.lines.len() as f32
	}

	pub fn update_width(&mut self, width: f32) {
		if self.max_width == width && !self.lines.is_empty() {
			return;
		}
		self.max_width = width;
		self.lines.clear();
		// todo fix splitting very long words
		let mut line_start = 0;
		let mut line_end = 0;
		for (i, c) in self.text.char_indices() {
			if c == ' ' {
				let line = &self.text[line_start..i];
				let new_line_length = measure_text(line, None, self.font_size as _, 1.).width;
				if new_line_length <= self.max_width {
					line_end = i;
				} else {
					self.lines.push(line_start..line_end);
					line_start = line_end;
				}
			}
			if c == '\n' {
				let line = &self.text[line_start..i];
				let new_line_length = measure_text(line, None, self.font_size as _, 1.).width;
				if new_line_length <= self.max_width {
					self.lines.push(line_start..i);
					line_end = i + 1;
					line_start = i + 1;
				} else {
					self.lines.push(line_start..line_end);
					self.lines.push(line_end..(i + 1));
					line_start = i + 1;
					line_end = i + 1;
				}
			}
		}
		let i = self.text.len();
		let line = &self.text[line_start..i];
		let new_line_length = measure_text(line, None, self.font_size as _, 1.).width;
		if new_line_length <= self.max_width {
			self.lines.push(line_start..i);
		} else {
			self.lines.push(line_start..line_end);
			self.lines.push(line_end..i);
		}
	}

	pub fn draw(&self, x: f32, y: f32) {
		// draw_rectangle(x, y, self.max_width, 4, Color::RED);
		for (i, line) in self.lines.iter().enumerate() {
			let line = &self.text[line.clone()];
			let line_y = y + self.font_size as f32 * i as f32;
			draw_text(line, x, line_y, self.font_size, colors::WHITE);
		}
	}
}

#[derive(Debug, Default)]
pub struct Tooltip {
	text: Option<&'static str>,
	mouse_x: f32,
	mouse_y: f32,
	screen_width: f32,
	screen_height: f32,
}

impl Tooltip {
	pub fn init_frame(&mut self) {
		let (mouse_x, mouse_y) = mouse_position();
		*self = Self {
			text: None,
			mouse_x,
			mouse_y,
			screen_width: screen_width(),
			screen_height: screen_height(),
		};
	}

	pub fn add(&mut self, x: f32, y: f32, width: f32, height: f32, text: &'static str) {
		if self.mouse_x >= x
			&& self.mouse_y >= y
			&& self.mouse_x <= (x + width)
			&& self.mouse_y <= (y + height)
		{
			self.text = Some(text);
		}
	}

	pub fn reset(&mut self) {
		self.text = None;
	}

	pub fn add_rec(&mut self, bounds: Rect, text: &'static str) {
		self.add(bounds.x, bounds.y, bounds.w, bounds.h, text);
	}

	pub fn draw(&self) {
		if let Some(text) = self.text {
			let font_size = 20.0;
			let margin = 4.0;
			let text_width = measure_text(text, None, font_size as _, 1.).width;
			let x = self
				.mouse_x
				.min(self.screen_width - text_width - margin * 2.);
			let y = self
				.mouse_y
				.min(self.screen_height - font_size - margin * 2.);
			draw_rectangle(
				x,
				y,
				text_width + margin * 2.,
				font_size + margin * 2.,
				BG_LIGHT,
			);
			draw_text(text, x + margin, y + margin, font_size, colors::WHITE);
		}
	}
}

pub fn toggle_button(mouse: &MouseInput, x: f32, y: f32, width: f32, height: f32, val: &mut bool) {
	let margin = 5.;
	let mouse_pos = mouse.pos();
	let bounds = Rect::new(x, y, width, height);

	let hover = bounds.contains(mouse_pos);
	draw_rectangle(x, y, width, height, widget_bg(hover));
	let pressed = hover && mouse.left_click();
	if pressed {
		*val = !*val;
	}
	if *val {
		draw_rectangle(
			x + margin,
			y + margin,
			width - margin * 2.,
			height - margin * 2.,
			colors::WHITE,
		);
	}
}

pub fn simple_button(mouse: &MouseInput, x: f32, y: f32, width: f32, height: f32) -> bool {
	let mouse_pos = mouse.pos();
	let bounds = Rect::new(x, y, width, height);
	let hover = bounds.contains(mouse_pos);
	let pressed = hover && mouse.left_click();
	draw_rectangle(x, y, width, height, widget_bg(hover));
	pressed
}

pub fn text_button(mouse: &MouseInput, x: f32, y: f32, width: f32, text: &str) -> bool {
	let font_size = 20.;
	let margin = font_size / 4.;
	let height = font_size + margin * 2.;
	let clicked = simple_button(mouse, x, y, width, height);
	draw_text(text, x + margin, y + margin, font_size, colors::WHITE);
	clicked
}

pub fn tex32_button(
	mouse: &MouseInput,
	(x, y): (f32, f32),
	texture: &Texture2D,
	(tooltip, text): (&mut Tooltip, &'static str),
) -> bool {
	let size = 32.;
	let clicked = simple_button(mouse, x, y, size, size);
	draw_scaled_texture(texture, x, y, 2.);
	tooltip.add(x, y, size, size, text);
	clicked
}

pub fn simple_option_button<T>(mouse: &MouseInput, bounds: Rect, option: T, current: &mut T) -> bool
where
	T: PartialEq,
{
	draw_rectangle_rec(bounds, widget_bg(&option == current));
	let mut changed = false;
	if mouse.left_click() && mouse.is_over(bounds) && current != &option {
		*current = option;
		changed = true;
	}
	changed
}

pub fn text_input(
	globals: &mut Globals,
	bounds: Rect,
	text: &mut String,
	is_selected: &mut bool,
	max_len: usize,
	editable: bool,
) -> bool {
	let mut changed = false;
	let font_size = 20.;
	draw_rectangle_rec(bounds, widget_bg(*is_selected));
	draw_rectangle(
		bounds.x + 2.,
		bounds.y + bounds.h - 5.,
		bounds.w - 4.,
		3.,
		BG_DARK,
	);
	draw_text(text, bounds.x + 4., bounds.y + 4., font_size, colors::WHITE);
	// blinking cursor
	if *is_selected && get_time().fract() < 0.5 {
		let width = measure_text(text, None, font_size as _, 1.).width;
		draw_rectangle(
			bounds.x + 6. + width,
			bounds.y + 4.,
			2.,
			font_size,
			colors::WHITE,
		);
	};
	if editable && globals.mouse.left_click() && (globals.mouse.is_over(bounds) || *is_selected) {
		*is_selected = !*is_selected;
	}

	if *is_selected {
		globals.config.input.in_text_edit = true;
		if is_key_pressed(KeyCode::Escape) || is_key_pressed(KeyCode::Enter) {
			*is_selected = false;
		}
		if is_key_pressed(KeyCode::Backspace) && !text.is_empty() {
			changed = true;
			text.pop();
			if is_key_down(KeyCode::LeftControl) {
				while let Some(c) = text.pop() {
					if c == ' ' {
						break;
					}
				}
			}
		}
		if text.len() < max_len {
			let c = get_char_pressed();
			if let Some(c) = c {
				changed = true;
				text.push(c);
			}
		}
	}
	changed
}

pub fn scrollable_texture_option_button<T>(
	mouse: &MouseInput,
	pos: Vec2,
	texture: &Texture2D,
	option: T,
	current: &mut T,
	border: f32,
	clicked_override: bool,
) -> Option<Scroll>
where
	T: PartialEq,
{
	let bounds = Rect::new(pos.x, pos.y, 32. + border * 2., 32. + border * 2.);
	draw_rectangle_rec(
		bounds,
		if &option == current {
			BG_WIDGET_ACTIVE
		} else {
			gray(16)
		},
	);
	let params = DrawTextureParams {
		dest_size: Some(Vec2::splat(32.)),
		..Default::default()
	};
	draw_texture_ex(
		texture,
		pos.x + border,
		pos.y + border,
		colors::WHITE,
		params,
	);
	if clicked_override {
		*current = option;
	} else if mouse.is_over(bounds) {
		if mouse.left_click() {
			*current = option;
		}
		return mouse.scroll();
	}
	None
}

pub fn draw_usize(
	textures: &Textures,
	mut number: usize,
	(x, y): (f32, f32),
	digits: u8,
	scale: f32,
) {
	for i in 0..digits {
		draw_rectangle(
			x + 10. * i as f32 * scale,
			y,
			8. * scale,
			16. * scale,
			BG_LIGHT,
		);
	}
	let mut i = 0;
	while (number != 0 || i == 0) && i < digits {
		let texture = textures.get(&format!("digit_{}", number % 10));
		let x = x + (digits - i - 1) as f32 * 10. * scale;
		draw_scaled_texture(texture, x, y, scale as f32);
		number /= 10;
		i += 1;
	}
}

pub fn draw_usize_small(textures: &Textures, mut num: usize, mut x: f32, y: f32, scale: f32) {
	const MAX_DIGITS: usize = 8;
	let mut digits = [0; MAX_DIGITS];
	let mut i = 0;
	while (num != 0 || i == 0) && i < MAX_DIGITS {
		digits[MAX_DIGITS - i - 1] = num % 10;
		num /= 10;
		i += 1;
	}
	let texture = textures.get("digits_small");
	for &digit in &digits[(MAX_DIGITS - i)..] {
		let source = Rect::new(4. * digit as f32, 0., 4., 6.);
		let params = DrawTextureParams {
			dest_size: Some(Vec2::new(4. * scale, 6. * scale)),
			source: Some(source),
			..Default::default()
		};
		draw_texture_ex(texture, x, y, FG_MARBLE_VALUE, params);
		x += 4. * scale;
	}
}

pub fn slider(mouse: &MouseInput, bounds: Rect, value: &mut u8, min: u8, max: u8) -> bool {
	// draw state
	// the +1 makes the lowest state look slightly filled and the max state fully filled
	let percent = (*value - min + 1) as f32 / (max - min + 1) as f32;
	draw_rectangle_rec(bounds, BG_WIDGET);
	let mut filled_bounds = bounds;
	filled_bounds.w *= percent;
	draw_rectangle_rec(filled_bounds, SLIDER_FILL);
	// interaction
	if mouse.is_over(bounds) {
		if mouse.left_hold() {
			let percent = (mouse.pos().x - bounds.x) / bounds.w;
			let new_value = min + (percent * (max - min + 1) as f32) as u8;
			if *value != new_value {
				*value = new_value;
			}
		} else if mouse.scroll() == Some(Scroll::Up) && *value < max {
			*value += 1;
		} else if mouse.scroll() == Some(Scroll::Down) && *value > min {
			*value -= 1;
		}
	}
	false
}
